indexing
   contents: "Memory access for moveable handles in PalmOS"
   author: "Victor Putz, vputz@nyx.net"
   completed: no
   tested: no

class MEM_HANDLE
   
inherit
   MEMORY
      redefine
	 dispose
      end
   
creation
   from_handle, allocate
   
feature {NONE} --internal pointer
   
   internal_locked_pointer : POINTER
   
   c_null_pointer : POINTER is
      external "C inline"
      alias "NULL"
      end
   
feature {ANY} -- state

   handle : POINTER

   is_locked : BOOLEAN
   
   should_release_on_destruction : BOOLEAN
   
   locked_pointer : POINTER is
      require
	 is_locked
      do
	 Result := internal_locked_pointer
      end

feature {NONE} -- creation

   from_handle( new_handle : like handle ) is
	 --initialize from an existing handle
      do
	 handle := new_handle;
	 is_locked := false;
	 internal_locked_pointer := c_null_pointer;
	 should_release_on_destruction := false
      end

   allocate( new_size : INTEGER_32; new_should_release_on_destruction : BOOLEAN ) is
	 --initialize by allocation
      do
	 handle := c_mem_handle_allocate( new_size )
	 is_locked := false;
	 internal_locked_pointer := c_null_pointer
	 should_release_on_destruction := new_should_release_on_destruction
      end
   
   dispose is
      do
	 if should_release_on_destruction then
	    release
	 end
      end

feature {ANY}

   card_number : INTEGER_16 is
      require
	 is_initialized
      do
	 Result := c_mem_handle_card_number( handle )
      end

   is_data_storage : BOOLEAN is
      require
	 is_initialized
      do
	 Result := c_mem_handle_is_data_storage( handle )
      end

   release is
	 --frees memory allocated by the handle.  YOU MUST DO THIS 
	 --DIRECTLY if should_free_on_destruction is false, which may 
	 --surprise people.  This is done to 
	 --prevent accidental memory deallocation.
      require
	 is_initialized
      local
	 err : INTEGER_16
      do
	 if is_locked then
	    unlock
	 end
	 if handle /= c_null_pointer then
	    err := c_mem_handle_free( handle )
	    handle := c_null_pointer
	 end
	 check 
	    err = 0
	 end
      ensure
	 not is_initialized
      end

   heap_id : INTEGER_16 is
      require
	 is_initialized
      do
	 Result := c_mem_handle_heap_id( handle )
      end

   lock is
      require
	 is_initialized
	 is_locked /= True
      do
	 internal_locked_pointer := c_mem_handle_lock( handle )
	 is_locked := True
      ensure
	 locked_pointer.is_not_null
	 is_locked = True
      end

   unlock is
      require
	 is_initialized
	 is_locked = True
      do
	 c_mem_handle_unlock( handle )
	 is_locked := False
	 internal_locked_pointer := c_null_pointer
      ensure
	 locked_pointer.is_null
	 is_locked = false
      end

   is_initialized : BOOLEAN is
      do
	 Result := handle.is_not_null
      end

   resize( new_size : INTEGER_32 ) is
      require
	 is_initialized
      local
	 err : INTEGER_16
      do
	 err := c_mem_handle_resize( handle, new_size )
	 check
	    err = 0
	 end
      ensure
	 size = new_size
      end

   set_owner( new_owner_id : INTEGER_16 ) is
      require
	 is_initialized
      local
	 err : INTEGER_16
      do
	 err := c_mem_handle_set_owner( handle, new_owner_id )
	 check
	    err = 0
	 end
      end

   size : INTEGER_32 is
      require
	 is_initialized
      do
	 Result := c_mem_handle_size( handle )
      end

   copy_from( p : POINTER; size_to_copy : INTEGER_32 ) is
      require
	 size > 0
      local
	 error : INTEGER_16
      do
	 if size < size_to_copy then
	    resize( size_to_copy )
	 end
	 lock
	 error := c_mem_move( locked_pointer, p, size_to_copy )
	 unlock
      end
   
   to_string : STRING is
	 --returns the contents of the handle as a string.
      local
	 h : MEM_HANDLE
	 --we use a temp handle to avoid lock nesting on this handle object
      do
	 create h.from_handle( handle )
	 if h.handle.is_not_null then
	    h.lock
	    create Result.from_external_copy( h.locked_pointer )
	    h.unlock
	 else
	    create Result.make(0)
	 end
      end
   
   to_local_id : LOCAL_ID is
      do
	 Result.set_value( c_mem_handle_to_local_id( handle ) )
      end
   
   set_to_string( s : STRING ) is
      require
	 s /= Void
      do
	 copy_from( s.to_external, s.count + 1 )
      end
   
feature {NONE} -- C implementation functions
   
   c_mem_handle_to_local_id( c_handle : like handle ) : POINTER is
      external "C use <MemoryMgr.h>"
      alias "MemHandleToLocalID"
      end
   
   c_mem_handle_allocate( new_size : INTEGER_32 ) : like handle is
      external "C use <MemoryMgr.h>"
      alias "MemHandleNew"
      end

   c_mem_handle_card_number( c_handle : like handle ) : INTEGER_16 is
      external "C use <MemoryMgr.h>"
      alias "MemHandleCardNo"
      end

   c_mem_handle_is_data_storage( c_handle : like handle ) : BOOLEAN is
      external "C use <MemoryMgr.h>"
      alias "MemHandleDataStorage"
      end

   c_mem_handle_free( c_handle : like handle ) : INTEGER_16 is 
	 --frees the handle.  Returns 0 if no error, or memErrInvalidParam
	 --if an error occurs.  TODO: add constant for memErrInvalidParam
      external "C use <MemoryMgr.h>" 
      alias "MemHandleFree"
      end

   c_mem_handle_heap_id( c_handle : like handle ) : INTEGER_16 is
      external "C use <MemoryMgr.h>"
      alias "MemHandleHeapID"
      end

   c_mem_handle_lock( c_handle : like handle ) : POINTER is
      external "C use <MemoryMgr.h>"
      alias "MemHandleLock"
      end

   c_mem_handle_unlock( c_handle : like handle ) is
      external "C use <MemoryMgr.h>"
      alias "MemHandleUnlock"
      end

   c_mem_handle_resize( c_handle : like handle; 
			new_size : INTEGER_32 ) : INTEGER_16 is
	 -- Resizes a handle.  Returns 0 for no error, or one of 
	 -- memErrInvalidParam, memErrNotEnoughSpace, memErrChunkLocked
	 -- TODO: add these constants
      external "C use <MemoryMgr.h>"
      alias "MemHandleResize"
      end

   c_mem_handle_set_owner( c_handle : like handle; 
			   new_owner_id : INTEGER_16 ) : INTEGER_16 is
	 --sets the new owner for the handle.  Returns 0 for no 
	 --error, or memErrInvalidParam if an error occurs
      external "C use <MemoryMgr.h>"
      alias "MemHandleSetOwner"
      end

   c_mem_handle_size( c_handle : like handle ) : INTEGER_32 is
      external "C use <MemoryMgr.h>"
      alias "MemHandleSize"
      end
   
   c_mem_move( dest, source : POINTER;
	       num_bytes : INTEGER_32 ) : INTEGER_16 is
      external "C use <MemoryMgr.h>"
      alias "MemMove"
      end
   

end
