module Pdbtools

  class Pdb_record_header
    attr_accessor :offset, :flags, :unique_id

    def initialize
      @offset = 0
      @flags = 0
      @unique_id = ""
    end

    def pack
      [ @offset, @flags, @unique_id ].pack( "NCa3" )
    end

    def unpack( str )
      ( @offset, @flags, @unique_id ) = str.unpack( "NCa3" )
    end

  end

  class Pdb

    attr_accessor :name, :attributes, :version, :create_time,
      :modify_time, :backup_time, :modification_number,
      :app_info_id, :sort_info_id, :type, :creator, :id_seed,
      :next_record_list

    Read_only_attribute = 0x02
    Dirty_appinfo_area_attribute = 0x04
    Backup_this_database_attribute = 0x08
    Okay_to_install_newer_attribute = 0x10
    Reset_after_install_attribute = 0x20
    Dont_allow_beaming_attribute = 0x40

    def time_from_rubytime( rubytime )
      rubytime.to_i + 2081376000
    end

    def initialize
      @name = ""
      @attributes = 0
      @version = 0
      @create_time = 0
      @modify_time = 0
      @backup_time = 0
      @modification_number = 0
      @app_info_id = 0
      @sort_info_id = 0
      @type = ""
      @creator = ""
      @id_seed = 0
      @next_record_list = 0
      @record_headers=[]
      @records = []
      @app_info = ""
      @sort_info = ""
    end

    def num_records 
      @records.length
    end

    def add_record( new_record )
      @records << new_record
    end

    def pack_header
      [ @name, @attributes, @version, @create_time,
	@modify_time, @backup_time, @modification_number,
	@app_info_id, @sort_info_id, @type, @creator, 
	@id_seed, @next_record_list, 
	num_records ].pack( "a32nnNNNNNNa4a4NNn" )
    end

    def unpack_partial_header( header )
      ( @name, @attributes, @version, @create_time,
       @modify_time, @backup_time, @modification_number,
       @app_info_id, @sort_info_id, @type, @creator, 
       @id_seed, @next_record_list ) = header.unpack( "a32nnNNNNNNa4a4NN" )
    end

    Record_header_size = 8
    Pdb_header_size = 78

    def writeable_app_info_offset
      if @app_info.length > 0
	Pdb_header_size + Record_header_size * num_records + 2
      else
	0
      end
    end

    def writeable_sort_info_offset
      if @sort_info.length > 0 
	app_info_offset + @app_info.size
      else
	0
      end
    end

    def writeable_first_record_offset
      if writeable_sort_info_offset != 0
	writeable_sort_info_offset + @sort_info.size
      elsif writeable_app_info_offset != 0
	writeable_app_info_offset + @app_info.size
      else
	Pdb_header_size + Record_header_size * num_records + 2
      end
    end

    def unpack_record_headers( num_records_to_read, packed_data_string )
      while num_records_to_read > 0 
	r = Pdb_record_header.new
	r.unpack( packed_data_string )
	@record_headers << r
	num_records_to_read = num_records_to_read - 1
	packed_data_string = packed_data_string[ Record_header_size..-1 ]
      end
    end

    def record_length( index, file_length ) 
      if index < @record_headers.length
	if index < (@record_headers.length - 1 )
	  @record_headers[index + 1].offset - @record_headers[ index ].offset
	else
	  file_length - @record_headers[ index ].offset
	end
      else
	-1
      end
    end

    def read_data_records( f ) 
      i = 0
      until i == @record_headers.length
	f.seek( @record_headers[i].offset )
	t = f.read( record_length (i, f.stat.size ) )
	@records << t
	p t
	i = i + 1
      end
    end

    def read_from_file( filename )
      #read the header
      initialize
      File.open( filename, "r" ) { |f|
	h = f.read( 76 )
	unpack_partial_header( h )
	#now read the record list
	h = f.read(2)
	num_records_to_read = h.unpack( "n" )[0]
	h = f.read( num_records_to_read * Record_header_size )
	unpack_record_headers( num_records_to_read, h )
	#read the app info area
	if @app_info_id != 0
	  f.seek( @app_info_id )
	  if @sort_info_id != 0 
	    @app_info = f.read ( @sort_info_id - @app_info_id )
	  elsif @record_headers.size > 0 
	    @app_info = f.read( @record_headers[0].offset - @app_info_id )
	  else
	    @app_info = f.read( f.stat.size - @sort_info_id )
	  end
	else
	  @app_info = ""
	end
	#read the sort info area
	if @sort_info_id != 0
	  f.seek( @sort_info_id )
	  if ( @record_headers.size > 0 )
	    @sort_info = f.read ( @record_headers[0].offset - @sort_info_id )
	  else
	    @sort_info = f.read( f.stat.size - @sort_info_id )
	  end
	else
	  @sort_info = ""
	end
	#read the data records
	read_data_records( f )
      }
    end

    def regenerate_record_headers
      #regenerates the record headers based on the 
      #records in @records[].  Existing record headers
      #will be destroyed.
      @record_headers = []
      offset = writeable_first_record_offset
      id = 0
      @records.each { |record|
	r = Pdb_record_header.new
	r.offset = offset
	r.flags = 0
	r.unique_id = [ 0, id ].pack( "Ca2" )
	@record_headers << r
      }
    end

    def regenerate_record_header_offsets
      #just regenerates the offsets in the record headers
      offset = writeable_first_record_offset
      i = 0
      until i == @records.size
	@record_headers[i].offset = offset
	offset = offset + @records[i].length
	i = i + 1
      end
    end

    def add_record( record, flags, unique_id )
      r = Pdb_record_header.new
      if @records.size == 0
	r.offset = writeable_first_record_offset
      else
	r.offset = @record_headers[-1].offset + @records[-1].length
      end
      r.flags = flags
      r.unique_id = unique_id
      @records << record
      @record_headers << r
    end

    def write_to_file( filename )
      #writes the PDB to a given filename.  The records
      #must be sorted first; this can be done by varying the
      #order in which they are added with add_record, or
      #by sorting separately and regenerating the record headers
      #with "regerate_record_headers"
      File.open( filename, "w" ) { |f|
	@app_info_id = writeable_app_info_offset
	@sort_info_id = writeable_sort_info_offset
	#write header
	f.write( pack_header )
	#write record headers
	regenerate_record_header_offsets
	@record_headers.each { |header|
	  f.write( header.pack )
	}
	#write two bytes of padding
	f.write( "\000\000" )
	#write app info
	f.write( @app_info )
	#write sort info
	f.write( @sort_info )
	#write records
	@records.each { |record|
	  f.write( record )
	}
      }
      
      
    end

  end

end

